import * as THREE from '../build/three.module.js';
import Stats from './jsm/libs/stats.module.js';

class App {
    constructor() {
        this._initialize();
    }

    _initialize() {
        const domWebGL = document.createElement('div');
        document.body.appendChild(domWebGL);
        this._domWebGL = domWebGL;

        const renderer = new THREE.WebGLRenderer({ antialias: true });
        domWebGL.appendChild(renderer.domElement);
        //renderer.outputEncoding = THREE.sRGBEncoding;
        this._renderer = renderer;

        const scene = new THREE.Scene();
        scene.background = new THREE.Color(0x050505);
        this._scene = scene;

        this._setupCamera();
        //this._setupLight();
        this._setupModel();
        this._setupGUI();

        //this._setupControls();
        //this._setupHelper();
        
        window.onresize = this.resize.bind(this);

        this.resize();
    }

    _setupModel() {
        const circleGeometry = new THREE.CircleGeometry(1, 6);

        const geometry = new THREE.InstancedBufferGeometry();
        geometry.index = circleGeometry.index;
        geometry.attributes = circleGeometry.attributes;

        const particleCount = 75000;
        const translateArray = new Float32Array(particleCount * 3);
        for(let i=0, i3=0, l=particleCount; i<l; i++, i3+=3) {
            translateArray[i3 + 0] = Math.random() * 2 - 1;
            translateArray[i3 + 1] = Math.random() * 2 - 1;
            translateArray[i3 + 2] = Math.random() * 2 - 1;
        }

        geometry.setAttribute("translate", new THREE.InstancedBufferAttribute(translateArray, 3));
        const material = new THREE.RawShaderMaterial({
            uniforms: {
                "map": { value: new THREE.TextureLoader().load("textures/sprites/circle.png")},
                "time": { value: 0.0 }
            },
            vertexShader: `
                precision highp float;
                
                uniform mat4 modelViewMatrix;
                uniform mat4 projectionMatrix;
                uniform float time;
        
                attribute vec3 position;
                attribute vec2 uv;
                attribute vec3 translate;
        
                varying vec2 vUv;
                varying float vScale;
        
                void main() {
                    vec4 mvPosition = modelViewMatrix * vec4( translate, 1.0 );
                    vec3 trTime = vec3(translate.x + time,translate.y + time,translate.z + time);
                    float scale =  sin( trTime.x * 2.1 ) + sin( trTime.y * 3.2 ) + sin( trTime.z * 4.3 );
                    vScale = scale;
                    scale = scale * 10.0 + 10.0;
                    mvPosition.xyz += position * scale;
                    vUv = uv;
                    gl_Position = projectionMatrix * mvPosition;
                }    
            `,
            fragmentShader: `
                precision highp float;

                uniform sampler2D map;

                varying vec2 vUv;
                varying float vScale;

                // HSL to RGB Convertion helpers
                vec3 HUEtoRGB(float H){
                    H = mod(H,1.0);
                    float R = abs(H * 6.0 - 3.0) - 1.0;
                    float G = 2.0 - abs(H * 6.0 - 2.0);
                    float B = 2.0 - abs(H * 6.0 - 4.0);
                    return clamp(vec3(R,G,B),0.0,1.0);
                }

                vec3 HSLtoRGB(vec3 HSL){
                    vec3 RGB = HUEtoRGB(HSL.x);
                    float C = (1.0 - abs(2.0 * HSL.z - 1.0)) * HSL.y;
                    return (RGB - 0.5) * C + HSL.z;
                }

                void main() {
                    vec4 diffuseColor = texture2D( map, vUv );
                    gl_FragColor = vec4( diffuseColor.xyz * HSLtoRGB(vec3(vScale/5.0, 1.0, 0.5)), diffuseColor.w );

                    if ( diffuseColor.w < 0.7 ) discard;
                }
            `,
            depthTest: true,
            depthWrite: true
        });

        const mesh = new THREE.Mesh(geometry, material);
        mesh.scale.set(500, 500, 500);
        this._scene.add(mesh);

        this._material = material;
        this._mesh = mesh;
    }

    _setupHelper() {
        //.
    }

    _setupGUI() {
        const stats = new Stats();
        this._domWebGL.appendChild(stats.dom);

        this._stats = stats;
    }

    _setupCamera() {
        const camera = new THREE.PerspectiveCamera(90, window.innerWidth / window.innerHeight, 1, 5000);
        camera.position.z = 1500;

        this._camera = camera;
    }

    _setupControls() {
        //.
    }

    _setupLight() {
        //.
    }

    update() {
        const time = performance.now() * 0.0005;
        
        this._material.uniforms["time"].value = time;
        this._mesh.rotation.x = time * 0.2;
        this._mesh.rotation.y = time * 0.4;

        this._stats.update();
    }

    render() {
        requestAnimationFrame(this.render.bind(this));
        
        this.update();
        this._renderer.render(this._scene, this._camera);
    }

    resize() {
        const aspect = window.innerWidth / window.innerHeight;

        this._camera.aspect = aspect;
        this._camera.updateProjectionMatrix();

        this._renderer.setPixelRatio( window.devicePixelRatio );
        this._renderer.setSize(window.innerWidth, window.innerHeight);
    }
}

window.onload = function () {
    (new App()).render();
}